import gzip
import shutil
import jellyfish
import time
import urllib.request
import os

class Download:
    '''
    Utility functions for downloading data
    '''
    
    @staticmethod
    def getURLContent(url:str):
        with urllib.request.urlopen(url) as urlResponse:
            content = urlResponse.read().decode()
            return content

    @staticmethod
    def getFileContent(path:str):
        with open(path, "r") as file:
            content = file.read()
            return content

    @staticmethod
    def needsDownload(filePath:str,force:bool=False)->bool:
        '''
        check if a download of the given filePath is necessary that is the file
        does not exist has a size of zero or the download should be forced
        
        Args:
            filePath(str): the path of the file to be checked
            force(bool): True if the result should be forced to True
            
        Return:
            bool: True if  a download for this file needed
        '''
        if not os.path.isfile(filePath):
            result=True
        else:
            stats=os.stat(filePath)
            size=stats.st_size
            result=force or size==0
        return result

    @staticmethod
    def downloadBackupFile(url:str, fileName:str, targetDirectory:str, force:bool=False):
        '''
        Downloads from the given url the zip-file and extracts the file corresponding to the given fileName.

        Args:
            url: url linking to a downloadable gzip file
            fileName: Name of the file that should be extracted from gzip file
            targetDirectory(str): download the file this directory
            force (bool): True if the download should be forced

        Returns:
            Name of the extracted file with path to the backup directory
        '''
        extractTo = f"{targetDirectory}/{fileName}"
        # we might want to check whether a new version is available
        if Download.needsDownload(extractTo, force=force):
            if not os.path.isdir(targetDirectory):
                os.makedirs(targetDirectory)
            zipped = f"{extractTo}.gz"
            print(f"Downloading {zipped} from {url} ... this might take a few seconds")
            urllib.request.urlretrieve(url, zipped)
            print(f"Unzipping {extractTo} from {zipped}")
            with gzip.open(zipped, 'rb') as gzipped:
                with open(extractTo, 'wb') as unzipped:
                    shutil.copyfileobj(gzipped, unzipped)
                print("Extracting completed")
            if not os.path.isfile(extractTo):
                raise (f"could not extract {fileName} from {zipped}")
        return extractTo


class Profiler:
    '''
    simple profiler
    '''
    def __init__(self,msg,profile=True):
        '''
        construct me with the given msg and profile active flag
        
        Args:
            msg(str): the message to show if profiling is active
            profile(bool): True if messages should be shown
        '''
        self.msg=msg
        self.profile=profile
        self.starttime=time.time()
        if profile:
            print(f"Starting {msg} ...")
    
    def time(self,extraMsg=""):
        '''
        time the action and print if profile is active
        '''
        elapsed=time.time()-self.starttime
        if self.profile:
            print(f"{self.msg}{extraMsg} took {elapsed:5.1f} s")
        return elapsed
        
        
def remove_non_ascii(s):
    ''' 
    Remove non ascii chars from the given string 
    Args:
        s: 
            string: The string to remove chars from 
    Returns:
        string: The result string with non-ascii chars removed 
        
    Hat tip: http://stackoverflow.com/a/1342373/2367526    
    '''
    return "".join(i for i in s if ord(i) < 128)


def fuzzy_match(s1, s2, max_dist=.8):
    ''' 
    Fuzzy match the given two strings with the given maximum distance
    jellyfish jaro_winkler_similarity based on https://en.wikipedia.org/wiki/Jaro-Winkler_distance
    Args:
        s1: 
            string: First string 
        s2: 
            string: Second string 
        max_dist: 
            float: The distance - default: 0.8 
    Returns:
        True if the match is greater equals max_dist. Otherwise false
    '''
    return jellyfish.jaro_winkler_similarity(s1, s2) >= max_dist



'''
Created on 2020-09-23

@author: wf
'''
import re
from geograpy.utils import Profiler
from lodstorage.sparql import SPARQL

class Wikidata(object):
    '''
    Wikidata access
    '''

    def __init__(self, endpoint='https://query.wikidata.org/sparql',profile:bool=True):
        '''
        Constructor
        '''
        self.endpoint=endpoint
        self.profile=profile
        
    def query(self,msg,queryString:str,limit=None)->list:
        '''
        get the query result
        
        Args:
            msg(str): the profile message to display
            queryString(str): the query to execute
            
        Return:
            list: the list of dicts with the result
        '''
        profile=Profiler(msg,profile=self.profile)
        wd=SPARQL(self.endpoint)
        limitedQuery=queryString
        if limit is not None:
            limitedQuery=f"{queryString} LIMIT {limit}"
        results=wd.query(limitedQuery)
        lod=wd.asListOfDicts(results)
        for record in lod:
            for key in list(record.keys()):
                value=record[key]
                if isinstance(value,str):
                    if value.startswith("http://www.wikidata.org/"):
                        record[key]=self.getWikidataId(value)
                    if key.lower().endswith("coord"):
                        lat, lon = Wikidata.getCoordinateComponents(value)
                        record["lat"]=lat
                        record["lon"]=lon
                        record.pop(key)
            
                        
        profile.time(f"({len(lod)})")
        return lod
    
    def store2DB(self,lod,tableName:str,primaryKey:str=None,sqlDB=None):
        '''
        store the given list of dicts to the database
        
        Args:
            lod(list): the list of dicts
            tableName(str): the table name to use
            primaryKey(str): primary key (if any)
            sqlDB(SQLDB): target SQL database
        '''
        msg=f"Storing {tableName}"
        profile=Profiler(msg,profile=self.profile)
        entityInfo = sqlDB.createTable(lod, entityName=tableName, primaryKey=primaryKey, withDrop=True,sampleRecordCount=-1)
        sqlDB.store(lod, entityInfo, fixNone=True)
        profile.time()
        
    def getCountries(self,limit=None):
        '''
        get a list of countries
        
        `try query <https://query.wikidata.org/#%23%20get%20a%20list%20of%20countries%0A%23%20for%20geograpy3%20library%0A%23%20see%20https%3A%2F%2Fgithub.com%2Fsomnathrakshit%2Fgeograpy3%2Fissues%2F15%0APREFIX%20rdfs%3A%20%3Chttp%3A%2F%2Fwww.w3.org%2F2000%2F01%2Frdf-schema%23%3E%0APREFIX%20wd%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2F%3E%0APREFIX%20wdt%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fdirect%2F%3E%0APREFIX%20p%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2F%3E%0APREFIX%20ps%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fstatement%2F%3E%0APREFIX%20pq%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fqualifier%2F%3E%0A%23%20get%20City%20details%20with%20Country%0ASELECT%20DISTINCT%20%3Fcountry%20%3FcountryLabel%20%3FcountryIsoCode%20%3FcountryPopulation%20%3FcountryGDP_perCapita%20%3Fcoord%20%20WHERE%20%7B%0A%20%20%23%20instance%20of%20City%20Country%0A%20%20%3Fcountry%20wdt%3AP31%2Fwdt%3AP279%2a%20wd%3AQ3624078%20.%0A%20%20%23%20label%20for%20the%20country%0A%20%20%3Fcountry%20rdfs%3Alabel%20%3FcountryLabel%20filter%20%28lang%28%3FcountryLabel%29%20%3D%20%22en%22%29.%0A%20%20%23%20get%20the%20coordinates%0A%20%20%3Fcountry%20wdt%3AP625%20%3Fcoord.%0A%20%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP297%20ISO%203166-1%20alpha-2%20code%0A%20%20%3Fcountry%20wdt%3AP297%20%3FcountryIsoCode.%0A%20%20%23%20population%20of%20country%0A%20%20%3Fcountry%20wdt%3AP1082%20%3FcountryPopulation.%0A%20%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP2132%0A%20%20%23%20nonminal%20GDP%20per%20capita%0A%20%20%3Fcountry%20wdt%3AP2132%20%3FcountryGDP_perCapita.%0A%7D>`_
      
        '''
        queryString="""# get a list of countries
# for geograpy3 library
# see https://github.com/somnathrakshit/geograpy3/issues/15
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX p: <http://www.wikidata.org/prop/>
PREFIX ps: <http://www.wikidata.org/prop/statement/>
PREFIX pq: <http://www.wikidata.org/prop/qualifier/>
# get City details with Country
SELECT DISTINCT ?wikidataid ?name ?iso ?pop ?coord
WHERE {
  BIND (?countryQ AS ?wikidataid)

  # instance of Country
  # inverse path see https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service/query_optimization#Inverse_property_paths
  wd:Q6256 ^wdt:P279*/^wdt:P31 ?countryQ .
  
  # VALUES ?country { wd:Q55}.
  # label for the country
  ?countryQ rdfs:label ?name filter (lang(?name) = "en").
  # get the continent (s)
  #OPTIONAL {
  #  ?country wdt:P30 ?continent.
  #  ?continent rdfs:label ?continentLabel filter (lang(?continentLabel) = "en").
  #}
  # get the coordinates
  OPTIONAL { 
      ?countryQ wdt:P625 ?coord.
  } 
  # https://www.wikidata.org/wiki/Property:P297 ISO 3166-1 alpha-2 code
  ?countryQ wdt:P297 ?iso.
  # population of country   
  OPTIONAL
  { 
    SELECT ?countryQ (max(?countryPopulationValue) as ?pop)
    WHERE {
      ?countryQ wdt:P1082 ?countryPopulationValue
    } group by ?countryQ
  }
  # https://www.wikidata.org/wiki/Property:P2132
  # nominal GDP per capita
  # OPTIONAL { ?country wdt:P2132 ?countryGDP_perCapitaValue. }
}
ORDER BY ?iso"""
        msg="Getting countries from wikidata ETA 10s"
        countryList=self.query(msg, queryString,limit=limit)
        return countryList

    def getRegions(self,limit=None):
        '''
        get Regions from Wikidata
        
        `try query <https://query.wikidata.org/#%23%20get%20a%20list%20of%20regions%0A%23%20for%20geograpy3%20library%0A%23%20see%20https%3A%2F%2Fgithub.com%2Fsomnathrakshit%2Fgeograpy3%2Fissues%2F15%0APREFIX%20rdfs%3A%20%3Chttp%3A%2F%2Fwww.w3.org%2F2000%2F01%2Frdf-schema%23%3E%0APREFIX%20wd%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2F%3E%0APREFIX%20wdt%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fdirect%2F%3E%0APREFIX%20wikibase%3A%20%3Chttp%3A%2F%2Fwikiba.se%2Fontology%23%3E%0ASELECT%20%3Fcountry%20%3FcountryLabel%20%3FcountryIsoCode%20%3Fregion%20%3FregionIsoCode%20%3FregionLabel%20%3Fpopulation%20%3Flocation%0AWHERE%0A%7B%0A%20%20%23%20administrative%20unit%20of%20first%20order%0A%20%20%3Fregion%20wdt%3AP31%2Fwdt%3AP279%2a%20wd%3AQ10864048.%0A%20%20OPTIONAL%20%7B%0A%20%20%20%20%20%3Fregion%20rdfs%3Alabel%20%3FregionLabel%20filter%20%28lang%28%3FregionLabel%29%20%3D%20%22en%22%29.%0A%20%20%7D%0A%20%20%23%20filter%20historic%20regions%0A%20%20%23%20FILTER%20NOT%20EXISTS%20%7B%3Fregion%20wdt%3AP576%20%3Fend%7D%0A%20%20%23%20get%20the%20population%0A%20%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP1082%0A%20%20OPTIONAL%20%7B%20%3Fregion%20wdt%3AP1082%20%3Fpopulation.%20%7D%0A%20%20%23%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP297%0A%20%20OPTIONAL%20%7B%20%0A%20%20%20%20%3Fregion%20wdt%3AP17%20%3Fcountry.%0A%20%20%20%20%23%20label%20for%20the%20country%0A%20%20%20%20%3Fcountry%20rdfs%3Alabel%20%3FcountryLabel%20filter%20%28lang%28%3FcountryLabel%29%20%3D%20%22en%22%29.%0A%20%20%20%20%3Fcountry%20wdt%3AP297%20%3FcountryIsoCode.%20%0A%20%20%7D%0A%20%20%23%20isocode%20state%2Fprovince%0A%20%20%3Fregion%20wdt%3AP300%20%3FregionIsoCode.%0A%20%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP625%0A%20%20OPTIONAL%20%7B%20%3Fregion%20wdt%3AP625%20%3Flocation.%20%7D%0A%7D>`_
        '''
        queryString="""# get a list of regions
# for geograpy3 library
# see https://github.com/somnathrakshit/geograpy3/issues/15
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX wikibase: <http://wikiba.se/ontology#>
SELECT DISTINCT ?countryId (?regionQ as ?wikidataid) ?name ?iso ?pop ?coord
WHERE
{
  # administrative unit of first order
  ?regionQ wdt:P31/wdt:P279* wd:Q10864048.
  OPTIONAL {
     ?regionQ rdfs:label ?name filter (lang(?name) = "en").
  }
  # isocode state/province (mandatory - filters historic regions while at it ...)
  # filter historic regions
  # FILTER NOT EXISTS {?region wdt:P576 ?end}
  { 
    SELECT ?regionQ (max(?regionAlpha2) as ?iso) (max(?regionPopulationValue) as ?pop) (max(?locationValue) as ?coord)
    WHERE {
      ?regionQ wdt:P300 ?regionAlpha2.
      # get the population
      # https://www.wikidata.org/wiki/Property:P1082
      OPTIONAL {
        ?regionQ wdt:P1082 ?regionPopulationValue
      } 
      # get the location
      # https://www.wikidata.org/wiki/Property:P625
      OPTIONAL {
        ?regionQ wdt:P625 ?locationValue. 
       }
    } GROUP BY ?regionQ
  }
  # # https://www.wikidata.org/wiki/Property:P297
  OPTIONAL { 
    ?regionQ wdt:P17 ?countryId.
  }
} ORDER BY ?iso"""
        msg="Getting regions from wikidata ETA 15s"
        regionList=self.query(msg, queryString,limit=limit)
        return regionList

        
    def getCities(self,limit=1000000):
        '''
        get all human settlements as list of dict with duplicates for label, region, country ...
        '''
        queryString="""PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>

SELECT DISTINCT (?cityQ as ?wikidataid) ?city ?altLabel ?geoNameId ?gndId ?cityPopulation ?cityCoord ?regionId ?countryId
WHERE {
  # instance of human settlement https://www.wikidata.org/wiki/Q486972
  wd:Q486972 ^wdt:P279*/^wdt:P31 ?cityQ .
  # Values
  # VALUES  ?cityQ { wd:Q656 }
 
  # label of the City
  ?cityQ rdfs:label ?city filter (lang(?city) = "en").
  
  OPTIONAL {
     ?cityQ skos:altLabel ?altLabel .
     FILTER (lang(?altLabel) = "en")
  }
  
  # geoName Identifier
  OPTIONAL {
      ?cityQ wdt:P1566 ?geoNameId.
  }

  # GND-ID
  OPTIONAL { 
      ?cityQ wdt:P227 ?gndId. 
  }
  
  # population of city
  OPTIONAL { 
    SELECT ?cityQ (max(?cityPopulationValue) as ?cityPopulation)
    WHERE {
      ?cityQ wdt:P1082 ?cityPopulationValue
    } group by ?cityQ
  }
  
  OPTIONAL{
     ?cityQ wdt:P625 ?cityCoord .
  }
  
  # region this city belongs to
  OPTIONAL {
    ?cityQ wdt:P131 ?regionId .     
  }

  # country this city belongs to
  OPTIONAL {
      ?cityQ wdt:P17 ?countryId .
  }
  
}
"""
        msg="Getting cities (human settlements) from wikidata ETA 50 s"
        citiesList=self.query(msg, queryString,limit=limit)
        return citiesList
    
    def getCitiesForRegion(self,regionId,msg):
        '''
        get the cities for the given Region
        '''
        regionPath="?region ^wdt:P131/^wdt:P131/^wdt:P131 ?cityQ." if regionId in ["Q980","Q21"] else "?cityQ wdt:P131* ?region." 
        queryString="""# get cities by region for geograpy3
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX wd: <http://www.wikidata.org/entity/>

SELECT distinct (?cityQ as ?wikidataid) ?name ?geoNameId ?gndId ?regionId ?countryId ?pop ?coord WHERE { 
  VALUES ?hsType {
      wd:Q1549591 wd:Q3957 wd:Q5119 wd:Q15284 wd:Q62049 wd:Q515 wd:Q1637706 wd:Q1093829 wd:Q486972 wd:Q532
  }
  
  VALUES ?region {
         wd:%s
  }
  
  # region the city should be in
  %s
  
  # type of human settlement to try
  ?hsType ^wdt:P279*/^wdt:P31 ?cityQ.
  
  # label of the City
  ?cityQ rdfs:label ?name filter (lang(?name) = "en").
   
  # geoName Identifier
  OPTIONAL {
      ?cityQ wdt:P1566 ?geoNameId.
  }

  # GND-ID
  OPTIONAL { 
      ?cityQ wdt:P227 ?gndId. 
  }
  
  OPTIONAL{
     ?cityQ wdt:P625 ?coord .
  }
  
  # region this city belongs to
  OPTIONAL {
    ?cityQ wdt:P131 ?regionId .     
  }
  
  OPTIONAL {
     ?cityQ wdt:P1082 ?pop
  }

  # country this city belongs to
  OPTIONAL {
      ?cityQ wdt:P17 ?countryId .
  }
}""" % (regionId,regionPath)           
        regionCities=self.query(msg, queryString)
        return regionCities

    def getCityStates(self, limit=None):
        '''
        get city states from Wikidata

        `try query <https://query.wikidata.org/#%23%20get%20a%20list%20of%20city%20states%0A%23%20for%20geograpy3%20library%0APREFIX%20rdfs%3A%20%3Chttp%3A%2F%2Fwww.w3.org%2F2000%2F01%2Frdf-schema%23%3E%0APREFIX%20wd%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2F%3E%0APREFIX%20wdt%3A%20%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fdirect%2F%3E%0APREFIX%20wikibase%3A%20%3Chttp%3A%2F%2Fwikiba.se%2Fontology%23%3E%0ASELECT%20DISTINCT%20%3FcountryId%20%28%3FcityStateQ%20as%20%3Fwikidataid%29%20%3Fname%20%3Fiso%20%3Fpop%20%3Fcoord%0AWHERE%0A%7B%0A%20%20%23%20all%20citiy%20states%0A%20%20%3FcityStateQ%20wdt%3AP31%20wd%3AQ133442%20.%0A%20%20%3FcityStateQ%20rdfs%3Alabel%20%3Fname%20filter%20%28lang%28%3Fname%29%20%3D%20%22en%22%29.%0A%20%20%7B%20%0A%20%20%20%20SELECT%20%3FcityStateQ%20%28max%28%3FisoCode%29%20as%20%3Fiso%29%20%28max%28%3FpopulationValue%29%20as%20%3Fpop%29%20%28max%28%3FlocationValue%29%20as%20%3Fcoord%29%0A%20%20%20%20WHERE%20%7B%0A%20%20%20%20%20%20%3FcityStateQ%20wdt%3AP300%7Cwdt%3AP297%20%3FisoCode.%0A%20%20%20%20%20%20%23%20get%20the%20population%0A%20%20%20%20%20%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP1082%0A%20%20%20%20%20%20OPTIONAL%20%7B%0A%20%20%20%20%20%20%20%20%3FcityStateQ%20wdt%3AP1082%20%3FpopulationValue%0A%20%20%20%20%20%20%7D%20%0A%20%20%20%20%20%20%23%20get%20the%20location%0A%20%20%20%20%20%20%23%20https%3A%2F%2Fwww.wikidata.org%2Fwiki%2FProperty%3AP625%0A%20%20%20%20%20%20OPTIONAL%20%7B%0A%20%20%20%20%20%20%20%20%3FcityStateQ%20wdt%3AP625%20%3FlocationValue.%20%0A%20%20%20%20%20%20%20%7D%0A%20%20%20%20%7D%20GROUP%20BY%20%3FcityStateQ%0A%20%20%7D%0A%20%20OPTIONAL%20%7B%20%0A%20%20%20%20%3FcityStateQ%20wdt%3AP17%20%3FcountryId.%0A%20%20%7D%0A%7D%20ORDER%20BY%20%3Fiso>`_
        '''
        queryString = """# get a list of city states
# for geograpy3 library
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX wikibase: <http://wikiba.se/ontology#>
SELECT DISTINCT ?countryId (?cityStateQ as ?wikidataid) ?name ?iso ?pop ?coord
WHERE
{
  # all citiy states
  ?cityStateQ wdt:P31 wd:Q133442 .
  ?cityStateQ rdfs:label ?name filter (lang(?name) = "en").
  { 
    SELECT ?cityStateQ (max(?isoCode) as ?iso) (max(?populationValue) as ?pop) (max(?locationValue) as ?coord)
    WHERE {
      ?cityStateQ wdt:P300|wdt:P297 ?isoCode.
      # get the population
      # https://www.wikidata.org/wiki/Property:P1082
      OPTIONAL {
        ?cityStateQ wdt:P1082 ?populationValue
      } 
      # get the location
      # https://www.wikidata.org/wiki/Property:P625
      OPTIONAL {
        ?cityStateQ wdt:P625 ?locationValue. 
       }
    } GROUP BY ?cityStateQ
  }
  OPTIONAL { 
    ?cityStateQ wdt:P17 ?countryId.
  }
} ORDER BY ?iso"""
        msg = "Getting regions from wikidata ETA 15s"
        cityStateList = self.query(msg, queryString, limit=limit)
        return cityStateList

    @staticmethod
    def getCoordinateComponents(coordinate:str) -> (float, float):
        '''
        Converts the wikidata coordinate representation into its subcomponents longitude and latitude
        Example: 'Point(-118.25 35.05694444)' results in ('-118.25' '35.05694444')

        Args:
            coordinate: coordinate value in the format as returned by wikidata queries

        Returns:
            Returns the longitude and latitude of the given coordinate as separate values
        '''
        # https://stackoverflow.com/a/18237992/1497139
        floatRegex=r"[-+]?\d+([.,]\d*)?"
        regexp=fr"Point\((?P<lon>{floatRegex})\s+(?P<lat>{floatRegex})\)"
        cMatch=None
        if coordinate:
            try:
                cMatch = re.search(regexp, coordinate)
            except Exception as ex:
                # ignore
                pass
        if cMatch:
            latStr=cMatch.group("lat")
            lonStr=cMatch.group("lon")
            lat,lon=float(latStr.replace(",",".")),float(lonStr.replace(",","."))
            if lon>180:
                lon=lon-360
            return lat,lon
        else:
            # coordinate does not have the expected format
            return None, None

    @staticmethod
    def getWikidataId(wikidataURL:str):
        '''
        Extracts the wikidata id from the given wikidata URL

        Args:
            wikidataURL: wikidata URL the id should be extracted from

        Returns:
            The wikidata id if present in the given wikidata URL otherwise None
        '''

        # regex pattern taken from https://www.wikidata.org/wiki/Q43649390 and extended to also support property ids
        wikidataidMatch = re.search(r"[PQ][1-9]\d*", wikidataURL)
        if wikidataidMatch and wikidataidMatch.group(0):
            wikidataid = wikidataidMatch.group(0)
            return wikidataid
        else:
            return None

    @staticmethod
    def getValuesClause(varName:str, values, wikidataEntities:bool=True):
        '''
        generates the SPARQL value clause for the given variable name containing the given values
        Args:
            varName: variable name for the ValuesClause
            values: values for the clause
            wikidataEntities(bool): if true the wikidata prefix is added to the values otherwise it is expected taht the given values are proper IRIs

        Returns:
            str
        '''
        clauseValues=""
        if isinstance(values, list):
            for value in values:
                if wikidataEntities:
                    clauseValues+=f"wd:{value} "
                else:
                    clauseValues+=f"{value} "
        else:
            if wikidataEntities:
                clauseValues = f"wd:{values} "
            else:
                clauseValues = f"{values} "
        clause = "VALUES ?%s { %s }" %(varName, clauseValues)
        return clause


'''
The locator module allows to get detailed city 
information including the region and country of a city from a 
location string.

Examples for location strings are:

    Amsterdam, Netherlands
    Vienna, Austria
    Vienna, IL
    Paris - Texas
    Paris TX
    
the locator will lookup the cities and try to disambiguate the result based on the country or region information found.

The results in string representationa are:
    
    Amsterdam (NH(North Holland) - NL(Netherlands))
    Vienna (9(Vienna) - AT(Austria))
    Vienna (IL(Illinois) - US(United States))
    Paris (TX(Texas) - US(United States)) 
    Paris (TX(Texas) - US(United States))
    
Each city returned has a city.region and city.country attribute with the details of the city.
    

Created on 2020-09-18

@author: wf
'''
import os
import glob
import urllib
import re
import csv
import sys
import gzip
import shutil
import json
from pathlib import Path

from lodstorage.entity import EntityManager
from lodstorage.storageconfig import StorageConfig, StoreMode
from sklearn.neighbors import BallTree
from geograpy.wikidata import Wikidata
from lodstorage.sql import SQLDB
from geograpy.utils import remove_non_ascii
from geograpy import wikidata
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from lodstorage.jsonable import JSONAble
from math import radians, cos, sin, asin, sqrt
from geograpy.utils import Profiler, Download

class LocationManager(EntityManager):
    '''
    a list of locations
    '''
    
    def __init__(self, name:str, entityName:str, entityPluralName:str, listName:str=None, tableName:str=None,clazz=None, primaryKey:str=None, config:StorageConfig=None, handleInvalidListTypes=True, filterInvalidListTypes=False, debug=False):
        '''
        construct me

        Args:
            name(string): name of this LocationManager
            entityName(string): entityType to be managed e.g. Country
            entityPluralName(string): plural of the the entityType e.g. Countries
            listName(str): the name of the list to hold
            tableName(str): the name of the table to use
            config(StorageConfig): the configuration to be used if None a default configuration will be used
            handleInvalidListTypes(bool): True if invalidListTypes should be converted or filtered
            filterInvalidListTypes(bool): True if invalidListTypes should be deleted
            debug(boolean): override debug setting when default of config is used via config=None
        '''
        if config is None:
            config=LocationContext.getDefaultConfig()
        super().__init__(name=name,
                          entityName=entityName,
                          entityPluralName=entityPluralName,
                          listName=listName,
                          clazz=clazz,
                          tableName=tableName,
                          primaryKey=primaryKey,
                          config=config,
                          handleInvalidListTypes=handleInvalidListTypes,
                          filterInvalidListTypes=filterInvalidListTypes,
                          debug=debug)
        self.balltree = None
        self.locationByWikidataID={}
        if config is not None and config.mode==StoreMode.SQL:
            self.sqldb=self.getSQLDB(config.cacheFile)

    def getBallTuple(self, cache:bool=True):
        '''
        get the BallTuple=BallTree,validList of this location list
        
        Args:
            cache(bool): if True calculate and use a cached version otherwise recalculate on
            every call of this function
            
        Returns:
            BallTree,list: a sklearn.neighbors.BallTree for the given list of locations, list: the valid list of locations
            list: valid list of locations
        '''
        validList = []
        if self.balltree is None or not cache:
            coordinatesrad = []
            for location in self.getList():
                if location.lat and location.lon:
                    latlonrad = (radians(location.lat), radians(location.lon))
                    coordinatesrad.append(latlonrad)
                    validList.append(location)
            self.ballTuple = BallTree(coordinatesrad, metric='haversine'), validList
        return self.ballTuple
    
    def fromCache(self,force=False,getListOfDicts=None,sampleRecordCount=-1):
        '''
        get me from the cache
        '''
        super().fromCache(force, getListOfDicts, sampleRecordCount)
        self.locationByWikidataID={}
        for entry in self.getList():
            self.locationByWikidataID[entry.wikidataid]=entry

    def getLocationByID(self, wikidataID:str):
        '''
        Returns the location object that corresponds to the given location

        Args:
            wikidataID: wikidataid of the location that should be returned

        Returns:
            Location object
        '''
        location=None
        if wikidataID in self.locationByWikidataID:
            location=self.locationByWikidataID[wikidataID]
        return location
    
    def add(self,location):
        '''
        add the given location to me 
        
        Args:
            location(object): the location to be added and put in my hash map
        '''
        self.getList().append(location)
        if hasattr(location,"wikidataid"):
            self.locationByWikidataID[location.wikidataid]=location

    
    @staticmethod
    def getBackupDirectory():
        path = str(Path(Path.home(), ".geograpy3"))
        return path
    
    @classmethod
    def downloadBackupFileFromGitHub(cls,fileName:str, targetDirectory:str=None, force:bool=False):
        '''
        download the given fileName from the github data directory
        
        Args:
            fileName(str): the filename to download
            targetDirectory(str): download the file this directory
            force(bool): force the overwriting of the existent file
        
        Return:
            str: the local file
        '''
        # Data is downloaded from the github wiki - to modify the data clone the wiki
        # as documented in https://github.com/somnathrakshit/geograpy3/wiki
        # git clone https://github.com/somnathrakshit/geograpy3.wiki.git
        url = f"https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/{fileName}.gz"
        if targetDirectory is None:
            targetDirectory=LocationManager.getBackupDirectory()
        backupFile = Download.downloadBackupFile(url, fileName, targetDirectory, force)
        return backupFile

    def getByName(self, *names:str):
        '''
        Get locations matching given names
        Args:
            name: Name of the location

        Returns:
            Returns locations that match the given name
        '''
        query = f"SELECT * FROM {self.clazz.__name__}Lookup WHERE label IN ({','.join('?'*len(names))})"
        sqldb=self.getSQLDB(self.config.cacheFile)
        locationRecords = sqldb.query(query, params=tuple(names))
        locations=self._locationsFromLookup(*locationRecords)
        return locations

    def getLocationsByWikidataId(self, *wikidataId:str):
        '''
        Returns Location objects for the given wikidataids
        Args:
            *wikidataId(str): wikidataIds of the locations that should be returned

        Returns:
            Location objects matching the given wikidataids
        '''
        wikidataIds=set(wikidataId)
        if wikidataIds is None or not wikidataIds:
            return
        query=f"SELECT * FROM {self.clazz.__name__}Lookup WHERE wikidataid IN ({','.join('?'*len(wikidataIds))})"
        sqldb = self.getSQLDB(self.config.cacheFile)
        locationRecords=sqldb.query(query, params=tuple(list(wikidataIds)))
        if locationRecords:
            locations=self._locationsFromLookup(*locationRecords)
            return locations
        else:
            if self.debug:
                print("No Records matching the given wikidataIds found.")
            return

    def _locationsFromLookup(self, *locationRecords:dict):
        '''
        Convert given lookup records to the corresponding location objects
        Args:
            *locationRecords: lookup records of locations

        Returns:
            List of Location objects based on the given records
        '''
        if self.clazz is City:
            locations=[City.fromCityLookup(record) for record in locationRecords]
        elif self.clazz is Region:
            locations = [Region.fromRegionLookup(record) for record in locationRecords]
        elif self.clazz is Country:
            locations = [Country.fromCountryLookup(record) for record in locationRecords]
        else:
            locations=[self.clazz.fromRecord(lr) for lr in locationRecords]
        return locations

    def getLocationByIsoCode(self, isoCode:str):
        '''
        Get possible locations matching the given isoCode
        Args:
            isoCode: isoCode of possible Locations

        Returns:
            List of wikidata ids of locations matching the given isoCode
        '''
        if isinstance(self, RegionManager) or isinstance(self, CountryManager):
            if isinstance(self, RegionManager):
                query = f"SELECT wikidataid FROM {self.tableName} WHERE iso LIKE (?) OR iso LIKE (?)"
                params = (f"%-{isoCode}", isoCode,)
            else:
                query = f"SELECT wikidataid FROM {self.tableName} WHERE iso LIKE (?)"
                params = (isoCode,)
            sqldb = self.getSQLDB(self.config.cacheFile)
            qres = sqldb.query(query, params)
            locationIds = [record['wikidataid'] for record in qres if 'wikidataid' in record]
            return locationIds
        else:
            return []

class CountryManager(LocationManager):
    '''
    a list of countries
    '''
    
    def __init__(self, name:str="CountryManager", config:StorageConfig=None, debug=False):
        super().__init__(name=name,
                            entityName="country",
                            entityPluralName="countries",
                            clazz=Country,
                            primaryKey="wikidataid",
                            tableName="countries",
                            config=config,
                            debug=debug
                            )
        self.wd=Wikidata()
        self.getListOfDicts=self.wd.getCountries

    @classmethod
    def fromErdem(cls):
        '''
        get country list provided by Erdem Ozkol https://github.com/erdem
        '''
        countryManager = CountryManager(name="countries_erdem")
        countryJsonUrl = "https://gist.githubusercontent.com/erdem/8c7d26765831d0f9a8c62f02782ae00d/raw/248037cd701af0a4957cce340dabb0fd04e38f4c/countries.json"
        with urllib.request.urlopen(countryJsonUrl) as url:
            jsonCountryList = json.loads(url.read().decode())
            for jsonCountry in jsonCountryList:
                country = Country()
                country.name = jsonCountry['name']
                country.iso = jsonCountry['country_code']
                country.lat = jsonCountry['latlng'][0]
                country.lon = jsonCountry['latlng'][1]
                countryManager.add(country)

        return countryManager


class RegionManager(LocationManager):
    '''
    a list of regions
    '''

    def __init__(self, name:str="RegionManager", config:StorageConfig=None,debug=False):
        super().__init__(name=name,
                            entityName="region",
                            entityPluralName="regions",
                            clazz=Region,
                            primaryKey="regionId",
                            tableName="regions",
                            config=config,
                            debug=debug
                        )
        self.wd=Wikidata()
        def _queryRegions(**kwargs):
            return [*self.wd.getRegions(**kwargs), *self.wd.getCityStates(**kwargs)]
        self.getListOfDicts=_queryRegions


class CityManager(LocationManager):
    '''
    a list of cities
    '''

    def __init__(self, name:str="CityManager",config:StorageConfig=None,debug=False):
        super().__init__(name=name,
                            entityName="city",
                            entityPluralName="cities",
                            clazz=City,
                            primaryKey=None,
                            tableName="cities",
                            config=config,
                            debug=debug
                        )
        self.wd=Wikidata()
        self.getListOfDicts=self.wd.getCities
       
    @classmethod    
    def getJsonFiles(cls,config:StorageConfig) -> list: 
        '''
        get the list of the json files that have my data
        
        Return:
            list: a list of json file names
        
        '''
        jsondir=f"{config.getCachePath()}/regions"
        if not os.path.exists(jsondir):
                os.makedirs(jsondir)
        jsonFiles = sorted(glob.glob(f"{jsondir}/*.json"), key=lambda path:int(re.findall(r'\d+', path)[0]))
        return jsonFiles
        
    
class Earth:
    radius = 6371.000  # radius of earth in km


class Location(JSONAble):
    '''
    Represents a Location
    '''

    def __init__(self, **kwargs):
        for key in kwargs.keys():
            setattr(self, key, kwargs[key])
        
    @classmethod
    def getSamples(cls):
        samplesLOD = [{
            "name": "Los Angeles",
            "wikidataid": "Q65",
            "lat": 34.05223,
            "lon":-118.24368,
            "partOf": "US/CA",
            "level": 5,
            "locationKind": "City",
            "comment": None,
            "population": 3976322
        }]
        return samplesLOD

    @staticmethod
    def haversine(lon1, lat1, lon2, lat2):
        """
        Calculate the great circle distance between two points 
        on the earth (specified in decimal degrees)
        """
        # convert decimal degrees to radians 
        lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    
        # haversine formula 
        dlon = lon2 - lon1 
        dlat = lat2 - lat1 
        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * asin(sqrt(a)) 
        return c * Earth.radius

    def getNClosestLocations(self, lookupLocationManager, n:int):
        """
        Gives a list of up to n locations which have the shortest distance to 
        me as calculated from the given listOfLocations
        
        Args:
            lookupLocationManager(LocationManager): a LocationManager object to use for lookup
            n(int): the maximum number of closest locations to return 
        
        Returns:
            list: a list of result Location/distance tuples
        """
        balltree, lookupListOfLocations = lookupLocationManager.getBallTuple()
        # check for n+1 entries since we might have my own record in the lookup list which we'll ignore late
        distances, indices = balltree.query([[radians(self.lat), radians(self.lon)]], k=n + 1, return_distance=True)
        resultLocations = self.balltreeQueryResultToLocationManager(distances[0], indices[0], lookupListOfLocations)
        return resultLocations
        
    def getLocationsWithinRadius(self, lookupLocationManager, radiusKm:float):
        """
        Gives the n closest locations to me from the given lookupListOfLocations
        
        Args:
            lookupLocationManager(LocationManager): a LocationManager object to use for lookup
            radiusKm(float): the radius in which to check (in km)
            
        Returns:
            list: a list of result Location/distance tuples
        """
        balltree, lookupListOfLocations = lookupLocationManager.getBallTuple()
        
        indices, distances = balltree.query_radius([[radians(self.lat), radians(self.lon)]], r=radiusKm / Earth.radius,
                                                    return_distance=True)
        locationList = self.balltreeQueryResultToLocationManager(distances[0], indices[0], lookupListOfLocations)
        return locationList
    
    def balltreeQueryResultToLocationManager(self, distances, indices, lookupListOfLocations):
        '''
        convert the given ballTree Query Result to a LocationManager
        
        Args:
            distances(list): array of distances
            indices(list): array of indices
            lookupListOfLocations(list): a list of valid locations to use for lookup
            
        Return:
            list: a list of result Location/distance tuples
        '''
        locationListWithDistance = []
        for i, locationIndex in enumerate(indices):
            distance = distances[i] * Earth.radius 
            location = lookupListOfLocations[locationIndex]
            # do not add myself or any other equivalent location
            if not distance < 0.0001:
                locationListWithDistance.append((location, distance))
        # sort by distance (Ball tree only does this for one of the queries ...)        
        locationListWithDistance = sorted(locationListWithDistance, key=lambda lwd: lwd[1])
        return locationListWithDistance

    def distance(self, other) -> float:
        '''
        calculate the distance to another Location
        
        Args:
            other(Location): the other location
            
        Returns:
            the haversine distance in km
        '''
        # see https://stackoverflow.com/questions/4913349/haversine-formula-in-python-bearing-and-distance-between-two-gps-points
        distance = Location.haversine(self.lon, self.lat, other.lon, other.lat)
        return distance

    def isKnownAs(self, name) -> bool:
        '''
        Checks if this location is known under the given name

        Args:
            name(str): name the location should be checked against

        Returns:
            True if the given name is either the name of the location or present in the labels of the location
        '''
        isKnown = False
        if hasattr(self, 'labels'):
            if name in self.labels:
                isKnown = True
        if hasattr(self, 'name'):
            if name == self.name:
                isKnown = True
        return isKnown

    @staticmethod
    def partialDict(record, clazz, keys=None):
        if keys is None:
            keys = clazz.getSamples()[0].keys()
        pDict = {k: v for k, v in record.items() if k in keys}
        return pDict

    @staticmethod
    def mappedDict(record, keyMapList: list):
        keyMap = {}
        for mkey, mValue in keyMapList:
            keyMap[mkey] = mValue
        pDict = {keyMap[k]: v for k, v in record.items() if k in keyMap.keys()}
        return pDict

    @classmethod
    def fromRecord(cls,regionRecord: dict):
        '''
        create  a location from a dict record

        Args:
            regionRecord(dict): the records as returned from a Query

        Returns:
            Region: the corresponding region information
        '''
        location=cls()
        location.fromDict(regionRecord)
        return location

    
class City(Location):
    '''
    a single city as an object
    '''

    def __init__(self, **kwargs):
        super(City, self).__init__(**kwargs)
        if not hasattr(self, 'level'):
            setattr(self, 'level', 5)
        if not hasattr(self, 'locationKind'):
            setattr(self, 'locationKind', "City")
        self._country = None
        self._region = None

    @classmethod
    def getSamples(cls):
        samplesLOD = [{
            "name": "Los Angeles",
            "wikidataid": "Q65",
            "lat": 34.05223,
            "lon":-118.24368,
            "geoNameId": "5368361",
            "gndId": "4036361-2",
            "partOf": "US/CA",
            "level": 5,
            "locationKind": "City",
            "pop": "3976322",
            "regionId": "Q99",
            "countryId": "Q30"
        }]
        return samplesLOD
    
    def __str__(self):
        name=self.name if hasattr(self,"name") else "?"
        text = f"{name} ({self.region} - {self.country})"
        return text

    @staticmethod
    def fromCityLookup(cityLookupRecord:dict):
        '''

        create a city from a cityLookupRecord and setting City, Region and Country while at it
        Args:
            cityRecord(dict): a map derived from the CityLookup view

        '''
        # we create city, region and country from scratch without true
        # object relational mapping and lookup from the locationContext 
        # this is only useful for small result sets that need no further interlinking
        city=City()
        # first take all params
        cityRecord=City.partialDict(cityLookupRecord,City)
        city.fromDict(cityRecord)

        regionRecord=City.mappedDict(cityLookupRecord,
            [("regionId","wikidataid"),("regionName","name"),("regionIso","iso"),("regionPop","pop"),("regionLat","lat"),("regionLon","lon")])
        city.region=Region.fromRecord(regionRecord)

        countryRecord=City.mappedDict(cityLookupRecord,
            [("countryId","wikidataid"),("countryName","name"),("countryIso","iso"),("countryLat","lat"),("countryLon","lon")])
        city.country=Country()
        city.country.fromDict(countryRecord)
        city.region.country=city.country
        return city
    
    def setValue(self, name, record):
        '''
        set a field value with the given name  to
        the given record dicts corresponding entry or none
        
        Args:
            name(string): the name of the field
            record(dict): the dict to get the value from
        '''
        if name in record:
            value = record[name]
        else:
            value = None
        setattr(self, name, value)

    @property
    def country(self):
        return self._country

    @country.setter
    def country(self, country):
        self._country = country

    @property
    def region(self):
        return self._region

    @region.setter
    def region(self, region):
        self._region = region

        
class Region(Location):
    '''
    a Region (Subdivision)
    '''

    def __init__(self, **kwargs):
        super(Region, self).__init__(**kwargs)
        if not hasattr(self, 'level'):
            setattr(self, 'level', 4)
        if not hasattr(self, 'locationKind'):
            setattr(self,'locationKind', "Region")
        self._country = None

    @classmethod
    def getSamples(cls):
        samplesLOD = [{
            "name": "California",
            "wikidataid": "Q99",
            "lat": 37.0,
            "lon":-120.0,
            "partOf": "US",
            "level": 4,
            "locationKind": "Region",
            "comment": None,
            "labels": ["CA", "California"],
            "iso": "US-CA",
            "country_wikidataid": "Q30"
        }]
        return samplesLOD
    
    def __str__(self):
        text = f"{self.iso}({self.name})" 
        return text

    @property
    def country(self):
        return self._country

    @country.setter
    def country(self, country):
        self._country = country

    @staticmethod
    def fromRegionLookup(regionLookupRecord: dict):
        '''

        create a region from a regionLookupRecord and setting Region and Country while at it
        Args:
            regionRecord(dict): a map derived from the CityLookup view
        '''
        # we create region and country from scratch without true
        # object relational mapping and lookup from the locationContext
        # this is only useful for small result sets that need no further interlinking
        region = Region()
        # first take all params
        regionRecord = Location.partialDict(regionLookupRecord, Region)
        region.fromDict(regionRecord)
        countryRecord = Location.mappedDict(regionLookupRecord,
                                        [("countryId", "wikidataid"), ("countryName", "name"), ("countryIso", "iso"),
                                         ("countryLat", "lat"), ("countryLon", "lon")])
        region.country = Country()
        region.country.fromDict(countryRecord)
        return region


class Country(Location):
    '''
    a country
    '''

    def __init__(self, lookupSource='sqlDB', **kwargs):
        '''
        coonstruct me
        '''
        super(Country, self).__init__(**kwargs)
        if not hasattr(self, 'level'):
            setattr(self, 'level', 3)
        if not hasattr(self, 'locationKind'):
            setattr(self, 'locationKind', "Country")

    @classmethod
    def getSamples(cls):
        samplesLOD = [
            {
            'wikidataid': 'Q38', 
             'name': 'Italy', 
             'iso': 'IT', 
             'pop': 60317000.0, 
             'lat': 42.5, 
             'lon': 12.5,
            },
            {
            "name": "United States of America",
            "wikidataid": "Q30",
            "lat": 39.82818,
            "lon":-98.5795,
            "partOf": "North America",
            "level": 3,
            "locationKind": "Country",
            "comment": None,
            "labels":["USA", "US", "United States of America"],
            "iso":"US"
        }, {
            }]
        return samplesLOD

    def __str__(self):
        text = f"{self.iso}({self.name})" 
        return text

    @staticmethod
    def fromCountryLookup(countryLookupRecord: dict):
        '''

        create a region from a regionLookupRecord and setting Region and Country while at it
        Args:
            regionRecord(dict): a map derived from the CityLookup view
        '''
        # we create region and country from scratch without true
        # object relational mapping and lookup from the locationContext
        # this is only useful for small result sets that need no further interlinking
        country = Country()
        countryRecord = Location.partialDict(countryLookupRecord, Region)
        country.fromDict(countryRecord)
        return country


class LocationContext(object):
    '''
    Holds LocationManagers of all hierarchy levels and provides methods to traverse through the levels
    '''
    db_filename="locations.db"

    def __init__(self, countryManager:CountryManager, regionManager:RegionManager, cityManager:CityManager, config:StorageConfig):
        '''
        construct me
        
        Args:
            countryManager(CountryManager): the country manager to be used
            regionManager(RegionManager): the region manager to be used
            cityManager(CityManager): the city manager to be used
        '''
        self.countryManager = countryManager
        self.regionManager = regionManager
        self.cityManager = cityManager
        self.locator=Locator(storageConfig=config)

    def interlinkLocations(self,warnOnDuplicates:bool=True,profile=True):
        '''
        Interlinks locations by adding the hierarchy references to the locations
        
        Args:
            warnOnDuplicates(bool): if there are duplicates warn 
        '''
        profile=Profiler("interlinking Locations", profile=profile) 
        duplicates=[]
        self._countryLookup, _dup = self.countryManager.getLookup("wikidataid")
        duplicates.extend(_dup)
        self._regionLookup, _dup = self.regionManager.getLookup("wikidataid")
        duplicates.extend(_dup)
        self._cityLookup, _dup = self.cityManager.getLookup("wikidataid")
        duplicates.extend(_dup)
        if len(duplicates)>0 and warnOnDuplicates:
            print(f"There are {len(duplicates)} duplicate wikidataids in the country,region and city managers used")
            if self.debug:
                print(duplicates)
        # interlink region with country
        for region in self.regions:
            country = self._countryLookup.get(getattr(region, 'countryId'))
            if country is not None and isinstance(country, Country):
                region.country = country

        # interlink city with region and country
        for city in self.cities:
            country = self._countryLookup.get(getattr(city, 'countryId'))
            if country is not None and isinstance(country, Country):
                city.country = country
            region = self._regionLookup.get(getattr(city, 'regionId'))
            if region is not None and isinstance(region, Region):
                city.region = region
        _elapsed=profile.time()
  
                
    def load(self,forceUpdate:bool=False,warnOnDuplicates:bool=False):
        '''
        load my data
        '''
        for manager in self.countryManager,self.regionManager,self.cityManager:
            manager.fromCache(force=forceUpdate)
        self.interlinkLocations(warnOnDuplicates=warnOnDuplicates)
    

    @classmethod
    def fromCache(cls, config:StorageConfig=None, forceUpdate:bool=False):
        '''
        Inits a LocationContext form Cache if existent otherwise init cache

        Args:
            config(StorageConfig): configuration of the cache if None the default config is used
            forceUpdate(bool): If True an existent cache will be over written
        '''
        if config is None:
            config = cls.getDefaultConfig()
        if Download.needsDownload(config.cacheFile):
            LocationManager.downloadBackupFileFromGitHub(fileName=cls.db_filename,
                                                         targetDirectory=config.getCachePath(),
                                                         force=forceUpdate)
        cityManager = CityManager("cities", config=config)
        regionManager = RegionManager("regions", config=config)
        countryManager = CountryManager("countries", config=config)
        locationContext = LocationContext(countryManager, regionManager, cityManager, config)
        return locationContext

    @staticmethod
    def getDefaultConfig() -> StorageConfig:
        '''
        Returns default StorageConfig
        '''
        config = StorageConfig(cacheFile=LocationContext.db_filename,cacheDirName="geograpy3")
        config.cacheFile=f"{config.getCachePath()}/{config.cacheFile}"
        return config

    @property
    def countries(self) -> list:
        return self.countryManager.getList()

    @property
    def regions(self) -> list:
        return self.regionManager.getList()

    @property
    def cities(self) -> list:
        return self.cityManager.getList()

    def locateLocation(self, *locations, verbose:bool=False):
        '''
        Get possible locations for the given location names.
        Current prioritization of the results is city(ordered by population)→region→country
        ToDo: Extend the ranking of the results e.g. matching of multiple location parts increase ranking
        Args:
            *locations:
            verbose(bool): If True combinations of locations names are used to improve the search results. (Increases lookup time)

        Returns:

        '''
        if locations is None or locations is (None):
            return
        locationParts = []
        for location in locations:
            if location is not None:
                for locationPart in location.split(','):
                    locationParts.append(locationPart)
        # Split locationParts even further
        lp=[]
        for locationPart in locationParts:
            parts=locationPart.split(' ')
            lp.extend(parts)
            # Spliting by space breakes the look up for cities such as 'Los Angeles'
            if verbose:
                numberParts=len(parts)
                if numberParts>1:
                    lp.extend([f"{parts[i]} {parts[i+1]}" for i in range(numberParts-1)])
                # if numberParts > 2:
                #     lp.extend([f"{parts[i]} {parts[i + 1]} {parts[i + 2]}" for i in range(numberParts - 2)])
        locationParts.extend(lp)
        locationParts=list(set(locationParts))   # remove duplicates

        cities=self.cityManager.getByName(*locationParts)
        regions = self.regionManager.getByName(*locationParts)
        countries = self.countryManager.getByName(*locationParts)

        # remove locations already identified by location in lower hierarchy
        getAttrValues=lambda locations, attr:[getattr(location,attr) for location in locations if hasattr(location, attr)]
        excludeRegionIds=getAttrValues(cities, 'regionId')
        regions=[region for region in regions if hasattr(region, 'wikidataid') and not region.wikidataid in excludeRegionIds]
        excludeCountryIds=[*getAttrValues(cities, "countryId"), *getAttrValues(regions, "countryId")]
        countries=[country for country in countries if hasattr(country, 'wikidataid') and not country.wikidataid in excludeCountryIds]

        # build final result in the order city→region→country
        cities.sort(key=lambda c: int(getattr(c, 'pop', 0)) if getattr(c, 'pop') is not None else 0, reverse=True)
        res = [*cities, *regions, *countries]
        return res



class Locator(object):
    '''
    location handling
    '''
    
    # singleton instance
    locator = None

    def __init__(self, db_file=None, correctMisspelling=False, storageConfig:StorageConfig=None, debug=False):
        '''
        Constructor
        
        Args:
            db_file(str): the path to the database file
            correctMispelling(bool): if True correct typical misspellings
            storageConfig(StorageConfig): the storage Configuration to use
            debug(bool): if True show debug information
        '''
        self.debug = debug
        self.correctMisspelling = correctMisspelling
        if storageConfig is None:
            storageConfig=LocationContext.getDefaultConfig()
        self.storageConfig=storageConfig
        if db_file is None:
            self.db_path = self.storageConfig.getCachePath()
            self.db_file = self.storageConfig.cacheFile
        else:
            self.db_file=db_file
        self.view = "CityLookup"
        self.loadDB()
        self.getAliases()
        self.dbVersion = "2021-08-18 16:15:00"
        
    @staticmethod
    def resetInstance():
        Locator.locator = None    
    
    @staticmethod
    def getInstance(correctMisspelling=False, debug=False):
        '''
        get the singleton instance of the Locator. If parameters are changed on further calls
        the initial parameters will still be in effect since the original instance will be returned!
        
        Args:
            correctMispelling(bool): if True correct typical misspellings
            debug(bool): if True show debug information
        '''
        if Locator.locator is None:
            Locator.locator = Locator(correctMisspelling=correctMisspelling, debug=debug)
        return Locator.locator

    def normalizePlaces(self,places:list):
        '''
        normalize places

        Args:
            places(list) a list of places

        Return:
            list: stripped and aliased list of places
        '''
        nplaces=[]
        for place in places:
            place = place.strip()
            if place in self.aliases:
                place = self.aliases[place]
            nplaces.append(place)
        return nplaces

    def locateCity(self, places:list):
        '''
        locate a city, region country combination based on the given wordtoken information
        
        Args:
            places(list): a list of places derived by splitting a locality e.g.  "San Francisco, CA"
            leads to "San Francisco", "CA"
        
        Returns:
            City: a city with country and region details
        '''
        # make sure the database is populated
        self.populate_db()
        country = None
        cities = []
        regions = []
        # loop over all word elements
        places=self.normalizePlaces(places)
        for place in places:
            foundCountry = self.getCountry(place)
            if foundCountry is not None:
                country = foundCountry
            foundCities = self.cities_for_name(place)
            cities.extend(foundCities)
            foundRegions = self.regions_for_name(place)
            regions.extend(foundRegions)
        foundCity = self.disambiguate(country, regions, cities)
        return foundCity

    @staticmethod
    def isISO(s):
        '''
        check if the given string is an ISO code (ISO 3166-2 code)
        see https://www.wikidata.org/wiki/Property:P300

        Returns:
            bool: True if the string might be an ISO Code as per a regexp check
        '''
        m = re.search(r"^([A-Z]{1,2}\-)?[0-9A-Z]{1,3}$", s)
        result = m is not None
        return result
               
    def disambiguate(self, country, regions, cities, byPopulation=True): 
        '''
        try determining country, regions and city from the potential choices
        
        Args:
            country(Country): a matching country found
            regions(list): a list of matching Regions found
            cities(list): a list of matching cities found
            
        Return:
            City: the found city or None
        '''
        if self.debug:
            print("countries: %s " % country)
            print("regions: %s" % "\n\t".join(str(r) for r in regions))
            print("cities: %s" % "\n\t".join(str(c) for c in cities))
        foundCity = None
        # is the city information unique?
        if len(cities) == 1:
            foundCity = cities[0]
        else: 
            if len(cities) > 1:
                if country is not None:
                    for city in cities:
                        if self.debug:
                            print("city %s: " % (city))
                        if city.country.iso == country.iso:
                            foundCity = city
                            break
                if foundCity is None and len(regions) > 0:
                    for region in regions:
                        for city in cities:
                            if city.region.iso == region.iso and not city.region.name == city.name:
                                foundCity = city
                                break
                        if foundCity is not None:
                            break
                if foundCity is None and byPopulation:
                    foundCity = max(cities, key=lambda city:0 if city.pop is None else city.pop)
                    pass
                    
        return foundCity    
    
    def cities_for_name(self, cityName):
        '''
        find cities with the given cityName
        
        Args:
            cityName(string): the potential name of a city
        
        Returns:
            a list of city records
        '''
        cities = []
        cityRecords = self.places_by_name(cityName, "name")
        for cityRecord in cityRecords:
            cities.append(City.fromCityLookup(cityRecord))
        return cities

    def regions_for_name(self, region_name):
        '''
        get the regions for the given region_name (which might be an ISO code)
        
        Args:
            region_name(string): region name
            
        Returns:
            list: the list of cities for this region
        '''
        regions = []    
        if self.isISO(region_name):
            columnName = "iso"
        else:
            columnName = 'name'
        query = f"SELECT * from regions WHERE {columnName} = (?)" 
        params = (region_name,)
        regionRecords = self.sqlDB.query(query, params)
        for regionRecord in regionRecords:
            regions.append(Region.fromRecord(regionRecord))
        return regions                     
    
    def correct_country_misspelling(self, name):
        '''
        correct potential misspellings 
        Args:
            name(string): the name of the country potentially misspelled
        Return:
            string: correct name of unchanged
        '''
        cur_dir = os.path.dirname(os.path.realpath(__file__))
        with open(cur_dir + "/data/ISO3166ErrorDictionary.csv") as info:
            reader = csv.reader(info)
            for row in reader:
                if name == remove_non_ascii(row[0]):
                    return row[2]
        return name

    def is_a_country(self, name):
        '''
        check if the given string name is a country
        
        Args:
            name(string): the string to check
        Returns:
            True: if pycountry thinks the string is a country
        '''
        country = self.getCountry(name)
        result = country is not None
        return result
       
    def getCountry(self, name):
        '''
        get the country for the given name    
        Args:
            name(string): the name of the country to lookup
        Returns:     
            country: the country if one was found or None if not
        '''
        if self.isISO(name):
            query="SELECT * FROM countries WHERE iso = (?)"""
            params=(name,)
        else:
            if self.correctMisspelling:
                name = self.correct_country_misspelling(name)
            query="""SELECT * FROM countries 
WHERE name LIKE (?)
OR wikidataid in (SELECT wikidataid FROM country_labels WHERE label LIKE (?))"""
            params=(name,name,)
        country = None
        self.populate_db()
        countryRecords=self.sqlDB.query(query,params)
        if len(countryRecords)==1:
            country=Country.fromRecord(countryRecords[0])
            pass
        return country
    
    def getView(self):
        '''
        get the view to be used
        
        Returns:
            str: the SQL view to be used for CityLookups e.g. CityLookup
        '''
        view = self.view
        return view
 
    def places_by_name(self, placeName, columnName):
        '''
        get places by name and column
        Args:
            placeName(string): the name of the place
            columnName(string): the column to look at
        '''
        if not self.db_has_data():
            self.populate_db()
        view = self.getView()
        query = f'SELECT * FROM {view} WHERE {columnName} = (?) ORDER BY pop DESC'
        params = (placeName,)
        cityLookupRecords = self.sqlDB.query(query, params)
        cityLookupRecords.sort(key=lambda cityRecord: float(cityRecord.get('pop')) if cityRecord.get('pop') is not None else 0.0,reverse=True)
        return cityLookupRecords
    
     
    def recreateDatabase(self):
        '''
        recreate my lookup database
        '''
        print(f"recreating database ... {self.db_file}")
        self.populate_db(force=True)
        
                
    def populate_db(self, force=False):
        '''
        populate the cities SQL database which caches the information from the GeoLite2-City-Locations.csv file
        
        Args:
            force(bool): if True force a recreation of the database
        '''
        hasData = self.db_has_data()
        if force:
            self.populate_Countries(self.sqlDB)
            self.populate_Regions(self.sqlDB)
            self.populate_Cities(self.sqlDB)
            self.createViews(self.sqlDB)
            self.populate_Version(self.sqlDB)
    
        elif not hasData:
            self.downloadDB()
        if not os.path.isfile(self.db_file):
            raise(f"could not create lookup database {self.db_file}")
        
    def downloadDB(self, forceUpdate:bool=False):
        '''
        download my database

        Args:
            forceUpdate(bool): force the overwriting of the existent file
        '''
        if Download.needsDownload(self.db_file) or forceUpdate:
            LocationManager.downloadBackupFileFromGitHub(fileName=LocationContext.db_filename,
                                                         targetDirectory=self.storageConfig.getCachePath(),
                                                         force=forceUpdate)
            self.loadDB()
        
            
    def populate_Version(self, sqlDB):
        '''
        populate the version table
        
        Args:
            sqlDB(SQLDB): target SQL database
        '''
        versionList = [{"version":self.dbVersion}]
        entityInfo = sqlDB.createTable(versionList, "Version", "version", withDrop=True)
        sqlDB.store(versionList, entityInfo)
        
    def readCSV(self, fileName:str):
        '''
        read the given CSV file
        
        Args:
            fileName(str): the filename to read
        
        '''
        records = []
        cur_dir = os.path.dirname(os.path.realpath(__file__))
        csvfile = f"{cur_dir}/data/{fileName}" 
        with open(csvfile) as info:
            reader = csv.DictReader(info)
            for row in reader:
                records.append(row)
        return records
        
    def getAliases(self):
        '''
        get the aliases hashTable
        '''
        aliases = self.readCSV("aliases.csv")
        self.aliases = {}
        for alias in aliases:
            self.aliases[alias['name']] = alias['alias']
        
           
    def populate_Countries(self, sqlDB):
        '''
        populate database with countries from wikiData
        
        Args:
            sqlDB(SQLDB): target SQL database
        '''
        wikidata = Wikidata()
        countryList=wikidata.getCountries()
        wikidata.store2DB(countryList, "countries",primaryKey=None,sqlDB=sqlDB)
 
    def populate_Regions(self, sqlDB):
        '''
        populate database with regions from wikiData
        
        Args:
            sqlDB(SQLDB): target SQL database
        '''
        wikidata = Wikidata()
        regionList=wikidata.getRegions()
        wikidata.store2DB(regionList, "regions", primaryKey=None, sqlDB=sqlDB)
   
    def populate_Cities(self, sqlDB):
        '''
        populate the given sqlDB with the Wikidata Cities
        
        Args:
            sqlDB(SQLDB): target SQL database
        '''
        #wikidata = Wikidata()
        #wikidata.endpoint="https://confident.dbis.rwth-aachen.de/jena/wdhs/sparql"
        #cityList=wikidata.getCities()
        #wikidata.store2DB(cityList, "cities",primaryKey=None,sqlDB=sqlDB)
        config=LocationContext.getDefaultConfig()
        regionManager = RegionManager(config=config)
        regionManager.fromCache()
        regionByIso,_dup=regionManager.getLookup("iso")
        jsonFiles=CityManager.getJsonFiles(config)
        msg=f"reading {len(jsonFiles)} cached city by region JSON cache files"
        profiler=Profiler(msg)
        cityManager=CityManager(config=config)
        cityManager.getList().clear()
        for jsonFileName in jsonFiles:
            isoMatch = re.search(r"/([^\/]*)\.json", jsonFileName)
            if not isoMatch:
                print(f"{jsonFileName} - does not match a known region's ISO code")
            else:
                rIso=isoMatch.group(1)
                region=regionByIso[rIso]
                with open(jsonFileName) as jsonFile:
                    cities4Region = json.load(jsonFile)
                    for city4Region in cities4Region:
                        city=City()
                        city.fromDict(city4Region)
                        if hasattr(city, "regionId"):
                            city.partOfRegionId=city.regionId
                        city.regionId=region.wikidataid
                        cityManager.add(city)
                        pass
        cityManager.store()
        profiler.time()
        
    def createViews(self, sqlDB):
        viewDDLs = ["DROP VIEW IF EXISTS CityLookup", """
CREATE VIEW CityLookup AS
SELECT 
   cl.label,
   ci.*,
   r.name as regionName ,r.iso as regionIso ,r.pop as regionPop,r.lat as regionLat, r.lon as regionLon,
   c.name as countryName,c.iso as countryIso,c.lat as CountryLat, c.lon as CountryLon
FROM 
city_labels cl
JOIN cities ci on ci.wikidataid=cl.wikidataid
JOIN regions r on ci.regionId=r.wikidataid
JOIN countries c on ci.countryId=c.wikidataid
""","DROP VIEW IF EXISTS RegionLookup",
"""CREATE VIEW RegionLookup AS
SELECT 
   rl.label,
   r.*,
   c.name as countryName,c.iso as countryIso,c.lat as CountryLat, c.lon as CountryLon
FROM 
region_labels rl
JOIN regions r on rl.wikidataid=r.wikidataid
JOIN countries c on r.countryId=c.wikidataid
""","DROP VIEW IF EXISTS CountryLookup",
"""CREATE VIEW CountryLookup AS
SELECT 
   cl.label,
   c.*
FROM 
country_labels cl
JOIN countries c on cl.wikidataid=c.wikidataid
""",
"DROP INDEX if EXISTS cityLabelByWikidataid",
"CREATE INDEX cityLabelByWikidataid ON city_labels (wikidataid)",
"DROP INDEX if EXISTS cityByWikidataid",
"CREATE INDEX cityByWikidataid ON cities (wikidataid)",
"DROP INDEX IF EXISTS cityByRegion",
"CREATE INDEX cityByRegion ON cities (regionId)",
"DROP INDEX IF EXISTS regionByCountry",
"CREATE INDEX regionByCountry ON regions (countryId)"]
        for viewDDL in viewDDLs:
            sqlDB.execute(viewDDL)
    
    def db_recordCount(self, tableList, tableName):
        '''
        count the number of records for the given tableName
        
        Args:
            tableList(list): the list of table to check
            tableName(str): the name of the table to check
            
        Returns
            int: the number of records found for the table 
        '''
        tableFound = False
        for table in tableList:
            if table['name'] == tableName:
                tableFound = True
                break
        count = 0
        if tableFound: 
            query = "SELECT Count(*) AS count FROM %s" % tableName
            countResult = self.sqlDB.query(query)
            count = countResult[0]['count']
        return count
     
    def db_has_data(self):
        '''
        check whether the database has data / is populated
        
        Returns:
            boolean: True if the cities table exists and has more than one record
        '''
        tableList = self.sqlDB.getTableList()
        hasCities = self.db_recordCount(tableList,"cities")>200000
        hasCountries = self.db_recordCount(tableList, "countries") > 200
        hasRegions = self.db_recordCount(tableList, "regions") > 3000
        hasVersion = self.db_recordCount(tableList, "Version") == 1
        versionOk = False
        if hasVersion:
            query = "SELECT version from Version"
            dbVersionList = self.sqlDB.query(query)
            versionOk = dbVersionList[0]['version'] == self.dbVersion
        # hasWikidataCities=self.db_recordCount(tableList,'City_wikidata')>100000
        ok = hasVersion and versionOk and hasCities and hasRegions and hasCountries
        return ok

    def loadDB(self):
        '''
        loads the database from cache and sets it as sqlDB property
        '''
        self.sqlDB = SQLDB(self.db_file, errorDebug=True)

    
__version__ = '0.2.2'
__date__ = '2020-09-26'
__updated__ = '2021-11-29'

DEBUG = 1

    
def main(argv=None):  # IGNORE:C0111
    '''main program.'''

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)    
        
    program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
    user_name = "Wolfgang Fahl"
    program_license = '''%s

  Created by %s on %s.
  Copyright 2020-2021 Wolfgang Fahl. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, user_name, str(__date__))

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-d", "--debug", dest="debug", action="store_true", help="if True show debug information")
        parser.add_argument("-cm", "--correctSpelling", dest="correctMisspelling", action="store_true", help="if True correct typical misspellings")
        parser.add_argument("-db", "--recreateDatabase", dest='recreateDatabase', action="store_true", help="recreate the database")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)

        # Process arguments
        args = parser.parse_args()
        loc = Locator.getInstance(correctMisspelling=args.correctMisspelling, debug=args.debug)
        if args.recreateDatabase:
            loc.recreateDatabase()
        else:
            print ("no other functionality yet ...")
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2     

        
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())        




from .utils import remove_non_ascii, fuzzy_match
from collections import Counter
from geograpy.locator import Locator, City, Region

"""
Takes a list of place names and works place designation (country, region, etc) 
and relationships between places (city is inside region is inside country, etc)
"""

class PlaceContext(Locator):
    '''
    Adds context information to a place name
    '''

    def __init__(self, place_names:list, setAll:bool=True,correctMisspelling:bool=False):
        '''
        Constructor
        
        Args:
            place_names:
                list: The place names to check
            setAll:
                boolean: True if all context information should immediately be set
            db_file:
                    string: Path to the database file to be used - if None the default "locs.db" will be used
        '''
        super().__init__()
        self.correctMisspelling=correctMisspelling
        self.places = self.normalizePlaces(place_names)
        if setAll:
            self.setAll()
            
    def __str__(self):
        '''
        Return a string representation of me
        '''
        text= "countries=%s\nregions=%s\ncities=%s\nother=%s" % (self.countries,self.regions,self.cities,self.other)
        return text

    def getRegions(self, countryName:str)->list:
        '''
        get a list of regions for the given countryName

        countryName(str): the countryName to check
        '''
        regions = []
        queryString="""SELECT r.*  FROM 
COUNTRIES c 
JOIN regions r ON r.countryId=c.wikidataid
WHERE c.name=(?)"""
        params=(countryName,)
        regionRecords=self.sqlDB.query(queryString, params)
        for regionRecord in regionRecords:
            region=Region.fromRecord(regionRecord)
            regions.append(region)
        return regions

    def get_region_names(self, countryName:str)->list:
        '''
        get region names for the given country
        
        Args:
            countryName(str): the name of the country
        '''
        if self.correctMisspelling:
            countryName = self.correct_country_misspelling(countryName)
        regionOfCountryQuery="""SELECT name 
        FROM regions 
        WHERE countryId IN (
            SELECT wikidataid 
            FROM countries
            WHERE name LIKE (?)
            OR wikidataid IN (
                SELECT wikidataid 
                FROM country_labels 
                WHERE label LIKE (?)
            )
        )"""
        regionRecords=self.sqlDB.query(regionOfCountryQuery, params=(countryName,countryName,))
        return [r.get('name') for r in regionRecords]

    def setAll(self):
        '''
        Set all context information
        '''
        self.set_countries()
        self.set_regions()
        self.set_cities()
        self.set_other()
        
    def set_countries(self):
        '''
        get the country information from my places
        '''
        countries = []
        for place in self.places:
            country=self.getCountry(place)
            if country is not None:
                countries.append(country.name)

        self.country_mentions = Counter(countries).most_common()
        self.countries = list(set(countries))
        pass

    def set_regions(self):
        '''
        get the region information from my places (limited to the already identified countries)
        '''
        regions = []
        self.country_regions = {}
        region_names = {}

        if not self.countries:
            self.set_countries()

        def region_match(place_name:str, region_name:str)->bool:
            '''
            Tests the similarity of the given strings after removing non ascii characters.
            Args:
                place_name(str): Place name
                region_name(str): valid region name to test against

            Returns:
                True if the similarity of both values is greater equals 80%. Otherwise False
            '''
            return fuzzy_match(remove_non_ascii(place_name),
                               remove_non_ascii(region_name))

        def is_region(place_name:str, region_names:list):
            '''
            Filters out the regions that are not similar to the given place_name
            Args:
                place_name(str): place name to check against the regions
                region_names(list): List of valid region names

            Returns:
                List of regions that are similar to the given place_name
            '''
            return any([region_match(place_name, rn) for rn in region_names])

        for country in self.countries:
            region_names = self.get_region_names(country)
            matched_regions = [p for p in set(self.places) if is_region(p, region_names)]

            regions += matched_regions
            self.country_regions[country] = list(set(matched_regions))

        self.region_mentions = Counter(regions).most_common()
        self.regions = list(set(regions))

    def set_cities(self):
        '''
        set the cities information
        '''
        self.cities = []
        self.country_cities = {}
        self.address_strings = []

        if not self.countries:
            self.set_countries()

        if not self.regions:
            self.set_regions()

        if not self.db_has_data():
            self.populate_db()
        # ToDo: Duplicate with Locator.city_for_name e.g. extend method to support multiple names
        placesWithoutDuplicates=set(self.places)
        params=",".join("?" * len(placesWithoutDuplicates))
        query="SELECT * FROM CityLookup WHERE name IN (" + params + ")"
        cityLookupRecords=self.sqlDB.query(query,list(placesWithoutDuplicates))
        cityLookupRecords.sort(key=lambda cityRecord: float(cityRecord.get('pop')) if cityRecord.get('pop') is not None else 0.0 , reverse=True)
        for cityLookupRecord in cityLookupRecords:
            city=City.fromCityLookup(cityLookupRecord)
           
            if city.name not in self.cities:
                self.cities.append(city.name)

            countryName=city.country.name
            if countryName not in self.countries:
                self.countries.append(countryName)
                self.country_mentions.append((countryName, 1))

            if countryName not in self.country_cities:
                self.country_cities[countryName] = []

            if city.name not in self.country_cities[countryName]:
                self.country_cities[countryName].append(city.name)
                regionName=city.region.name
                if countryName in self.country_regions and regionName in self.country_regions[countryName]:
                    address=f"{city.name}, {regionName}, {countryName}"
                    self.address_strings.append(address)

        all_cities = [p for p in self.places if p in self.cities]
        self.city_mentions = Counter(all_cities).most_common()

    def set_other(self):
        if not self.cities:
            self.set_cities()

        def unused(place_name):
            places = [self.countries, self.cities, self.regions]
            return all(self.correct_country_misspelling(place_name) not in l for l in places)

        self.other = [p for p in self.places if unused(p)]



import nltk
import re
from newspaper import Article
from geograpy.labels import Labels

class Extractor(object):
    '''
    Extract geo context for text or from url
    '''
    def __init__(self, text=None, url=None, debug=False):
        '''
        Constructor
        Args:

            text(string): the text to analyze
            url(string): the url to read the text to analyze from
            debug(boolean): if True show debug information
        '''
        if not text and not url:
            raise Exception('text or url is required')
        self.debug=debug
        self.text = text
        self.url = url
        self.places = []
        nltk_packages = ['maxent_ne_chunker',
                        'words',
                        'treebank',
                        'maxent_treebank_pos_tagger',
                        'punkt',
                        'averaged_perceptron_tagger'
                        ]
        for nltk_package in nltk_packages:
            try:
                import nltk
                nltk.data.find(nltk_package)
            except LookupError:
                nltk.downloader.download(nltk_package, quiet=True)
        import nltk  

    def set_text(self):
        '''
        Setter for text
        '''
        if not self.text and self.url:
            a = Article(self.url)
            a.download()
            a.parse()
            self.text = a.text
            
    def split(self,delimiter=r","):
        '''
        simpler regular expression splitter with not entity check
        
        hat tip: https://stackoverflow.com/a/1059601/1497139
        '''
        self.set_text()
        self.places=re.split(delimiter,self.text)
            
    def find_geoEntities(self):
        '''
        Find geographic entities
        
        Returns:
            list: 
                List of places
        '''
        self.find_entities(Labels.geo)
        return self.places
        
    def find_entities(self,labels=Labels.default):
        '''
        Find entities with the given labels set self.places and returns it
        Args:
            labels: 
                Labels: The labels to filter
        Returns:
            list: 
                List of places
        '''
        self.set_text()

        text = nltk.word_tokenize(self.text)
        nes = nltk.ne_chunk(nltk.pos_tag(text))

        for ne in nes:
            if type(ne) is nltk.tree.Tree:
                nelabel=ne.label()
                if (nelabel in labels):
                    leaves=ne.leaves()
                    if self.debug:
                        print(leaves)
                    self.places.append(u' '.join([i[0] for i in leaves]))
        return self.places


'''
main geograpy 3 module
'''
from geograpy.extraction import Extractor
from geograpy.places import PlaceContext
from geograpy.locator import Locator
from geograpy.labels import Labels

def get_geoPlace_context(url=None, text=None,debug=False):
    '''
    Get a place context for a given text with information
    about country, region, city and other
    based on NLTK Named Entities having the Geographic(GPE) label.
    
    Args:
        url(String): the url to read text from (if any)
        text(String): the text to analyze
        debug(boolean): if True show debug information
    
    Returns:
        places: 
            PlaceContext: the place context
    '''    
    places=get_place_context(url, text, labels=Labels.geo, debug=debug)
    return places
    
def get_place_context(url=None, text=None,labels=Labels.default, debug=False):
    '''
    Get a place context for a given text with information
    about country, region, city and other
    based on NLTK Named Entities in the label set Geographic(GPE), 
    Person(PERSON) and Organization(ORGANIZATION).
    
    Args:
        url(String): the url to read text from (if any)
        text(String): the text to analyze
        debug(boolean): if True show debug information
    
    Returns:
        pc: 
            PlaceContext: the place context
    '''
    e = Extractor(url=url, text=text,debug=debug)
    e.find_entities(labels=labels)
    places=e.places
    pc = PlaceContext(places)
    pc.setAll()
    return pc

def locateCity(location,correctMisspelling=False,debug=False):
    '''
    locate the given location string
    Args:
        location(string): the description of the location
    Returns:
        Locator: the location
    '''
    e = Extractor(text=location,debug=debug)
    e.split()
    loc=Locator.getInstance(correctMisspelling=correctMisspelling,debug=debug)
    city=loc.locateCity(e.places)
    return city
    



from setuptools import setup
import pathlib
import pkg_resources
import os 
from collections import OrderedDict
import configparser

config = configparser.ConfigParser()
config.read('settings.ini')

with pathlib.Path(os.path.join(os.path.dirname(__file__), 'requirements.txt')).open() as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements_txt)
    ]
        
try:
    long_description = ""
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()

except:
    print('Curr dir:', os.getcwd())
    long_description = open('../../README.md').read()

setup(name='geograpy3',
      version=config["DEFAULT"]["version"],
      description='Extract countries, regions and cities from a URL or text',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/somnathrakshit/geograpy3',
      download_url='https://github.com/somnathrakshit/geograpy3',
      author='Somnath Rakshit',
      author_email='somnath52@gmail.com',
      license='Apache',
      project_urls=OrderedDict(
        (
            ("Documentation", "https://geograpy3.readthedocs.io"),
            ("Code", "https://github.com/somnathrakshit/geograpy3"),
            ("Issue tracker", "https://github.com/somnathrakshit/geograpy3/issues"),
        )
      ),
      classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11'
      ],
      packages=['geograpy'],
      install_requires=install_requires,
      package_data={
          'geograpy': ['data/*.csv'],
      },
      zip_safe=False)
